"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const R = __importStar(require("ramda"));
const Rx = __importStar(require("rxjs"));
const cli_table_1 = __importDefault(require("cli-table"));
const commander_1 = __importDefault(require("commander"));
const types_1 = require("./types");
const _1 = require("./");
const protocol_1 = require("@melonproject/protocol");
const token_math_1 = require("@melonproject/token-math");
const operators_1 = require("rxjs/operators");
const debug = require('debug')('exchange-aggregator');
const style = {
    head: ['ID', 'Exchange', 'Price', 'Volume', 'Cummulative volume'],
    colWidths: [25, 15, 25, 25, 25],
    chars: {
        top: '═',
        'top-mid': '╤',
        'top-left': '╔',
        'top-right': '╗',
        bottom: '═',
        'bottom-mid': '╧',
        'bottom-left': '╚',
        'bottom-right': '╝',
        left: '║',
        'left-mid': '╟',
        mid: '─',
        'mid-mid': '┼',
        right: '║',
        'right-mid': '╢',
        middle: '│',
    },
};
const displayPrice = (price, decimals) => {
    const value = token_math_1.toFixed(price, decimals);
    const base = price.base.token.symbol;
    const quote = price.quote.token.symbol;
    return `${value} ${base}/${quote}`;
};
const displayVolume = (quantity) => {
    const base = quantity.token.symbol;
    const volume = token_math_1.toFixed(quantity);
    return `${volume} ${base}`;
};
const exchangeOrderObservableCreators = {
    [types_1.Exchange.RADAR_RELAY]: (options) => {
        return _1.exchanges.radarrelay.watch(options);
    },
    [types_1.Exchange.KRAKEN]: (options) => {
        return _1.exchanges.kraken.watch(options);
    },
    [types_1.Exchange.KYBER_NETWORK]: (options) => {
        return _1.exchanges.kyber.watch(options);
    },
    [types_1.Exchange.ETHFINEX]: (options) => {
        return _1.exchanges.ethfinex.watch(options);
    },
    [types_1.Exchange.OASIS_DEX]: (options) => {
        const endpoint = 'ws://localhost:8545';
        const environment = protocol_1.constructEnvironment({ endpoint });
        return _1.exchanges.oasisdex.watch(Object.assign({}, options, { environment }));
    },
};
const exchangeOrderFetcherCreators = {
    [types_1.Exchange.RADAR_RELAY]: (options) => {
        return _1.exchanges.radarrelay.fetch(options);
    },
    [types_1.Exchange.KRAKEN]: (options) => {
        return _1.exchanges.kraken.fetch(options);
    },
    [types_1.Exchange.KYBER_NETWORK]: (options) => {
        return _1.exchanges.kyber.fetch(options);
    },
    [types_1.Exchange.ETHFINEX]: (options) => {
        return _1.exchanges.ethfinex.fetch(options);
    },
    [types_1.Exchange.OASIS_DEX]: (options) => {
        const endpoint = 'ws://localhost:8545';
        const environment = protocol_1.constructEnvironment({ endpoint });
        return _1.exchanges.oasisdex.fetch(Object.assign({}, options, { environment }));
    },
};
const createExchangeOrderObservables = (options, exchanges) => {
    return exchanges.map(exchange => {
        return exchangeOrderObservableCreators[exchange](options);
    });
};
const createExchangeOrderFetchers = (options, exchanges) => {
    return exchanges.map(exchange => {
        return exchangeOrderFetcherCreators[exchange](options);
    });
};
const watch = (options, exchanges) => {
    debug('Aggregating orderbook for %s.', exchanges.join(', '));
    const observables = createExchangeOrderObservables(options, exchanges).map(observable$ => observable$.pipe(operators_1.catchError(() => Rx.empty())));
    const orderbook$ = Rx.merge(...observables).pipe(operators_1.scan(_1.reduceOrderEvents, { asks: [], bids: [] }), operators_1.tap(orderbook => {
        const bids = orderbook.bids.length;
        const asks = orderbook.asks.length;
        debug('Aggregated orderbook with %s bids and %s asks.', bids, asks);
    }), operators_1.throttleTime(5000));
    return new Promise((reject, resolve) => {
        orderbook$.subscribe(orderbook => {
            const bids = new cli_table_1.default(style);
            orderbook.bids.forEach(value => {
                const price = displayPrice(value.trade);
                const volume = displayVolume(value.trade.base);
                const cummulative = displayVolume(value.cummulative);
                bids.push([value.id, value.exchange, price, volume, cummulative]);
            });
            const asks = new cli_table_1.default(style);
            orderbook.asks.forEach(value => {
                const price = displayPrice(value.trade);
                const volume = displayVolume(value.trade.base);
                const cummulative = displayVolume(value.cummulative);
                asks.push([value.id, value.exchange, price, volume, cummulative]);
            });
            console.log('Asks');
            console.log(asks.toString());
            console.log('Bids');
            console.log(bids.toString());
        }, error => {
            reject(error);
        }, () => {
            resolve();
        });
    });
};
const fetch = (options, exchanges) => __awaiter(this, void 0, void 0, function* () {
    const promises = createExchangeOrderFetchers(options, exchanges);
    const results = yield Promise.all(promises.map(promise => promise.catch(error => {
        debug('Error: %s', error);
        return [];
    })));
    const orders = _1.aggregateOrders([].concat(...results));
    const orderbook = _1.createOrderbook(options, orders);
    const bids = new cli_table_1.default(style);
    orderbook.bids.forEach(value => {
        const price = displayPrice(value.trade);
        const volume = displayVolume(value.trade.base);
        const cummulative = displayVolume(value.cummulative);
        bids.push([value.id, value.exchange, price, volume, cummulative]);
    });
    const asks = new cli_table_1.default(style);
    orderbook.asks.forEach(value => {
        const price = displayPrice(value.trade);
        const volume = displayVolume(value.trade.base);
        const cummulative = displayVolume(value.cummulative);
        asks.push([value.id, value.exchange, price, volume, cummulative]);
    });
    console.log('Asks');
    console.log(asks.toString());
    console.log('Bids');
    console.log(bids.toString());
});
commander_1.default.description('The exchange-aggregator testing tool.');
commander_1.default
    .command('orders [exchanges...]')
    .option('-w, --watch', 'Wether or not to subscribe to updates automatically.')
    .option('-n, --network <network>', 'The network (KOVAN/MAINNET).', /(MAINNET|KOVAN)/, 'MAINNET')
    .option('-b, --base <symbol>', 'The base token symbol.', /[A-Z]{3,4}/, 'ZRX')
    .option('-q, --quote <symbol>', 'The quote token symbol.', /[A-Z]{3,4}/, 'ETH')
    .description('Retrieve orders from the given exchanges')
    .action((args, options) => __awaiter(this, void 0, void 0, function* () {
    const supported = Object.keys(types_1.Exchange);
    const invalid = R.difference(args, supported);
    if (invalid && invalid.length) {
        console.error(`Invalid exchanges %s. Supported exchanges are %s.`, invalid.join(', '), supported.join(', '));
        process.exit(1);
    }
    const exchanges = args.length ? args : supported;
    const opts = {
        network: types_1.Network[options.network],
        pair: {
            base: token_math_1.createToken(options.base),
            quote: token_math_1.createToken(options.quote),
        },
    };
    if (!!options.watch) {
        return watch(opts, exchanges);
    }
    return fetch(opts, exchanges);
}));
commander_1.default.parse(process.argv);
